/*
 *
 *    OPEN-XCHANGE legal information
 *
 *    All intellectual property rights in the Software are protected by
 *    international copyright laws.
 *
 *
 *    In some countries OX, OX Open-Xchange, open xchange and OXtender
 *    as well as the corresponding Logos OX Open-Xchange and OX are registered
 *    trademarks of the Open-Xchange, Inc. group of companies.
 *    The use of the Logos is not covered by the GNU General Public License.
 *    Instead, you are allowed to use these Logos according to the terms and
 *    conditions of the Creative Commons License, Version 2.5, Attribution,
 *    Non-commercial, ShareAlike, and the interpretation of the term
 *    Non-commercial applicable to the aforementioned license is published
 *    on the web site http://www.open-xchange.com/EN/legal/index.html.
 *
 *    Please make sure that third-party modules and libraries are used
 *    according to their respective licenses.
 *
 *    Any modifications to this package must retain all copyright notices
 *    of the original copyright holder(s) for the original code used.
 *
 *    After any such modifications, the original and derivative code shall remain
 *    under the copyright of the copyright holder(s) and/or original author(s)per
 *    the Attribution and Assignment Agreement that can be located at
 *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
 *    given Attribution for the derivative code and a license granting use.
 *
 *     Copyright (C) 2004-2012 Open-Xchange, Inc.
 *     Mail: info@open-xchange.com
 *
 *
 *     This program is free software; you can redistribute it and/or modify it
 *     under the terms of the GNU General Public License, Version 2 as published
 *     by the Free Software Foundation.
 *
 *     This program is distributed in the hope that it will be useful, but
 *     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *     for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc., 59
 *     Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package com.openexchange.office.ooxml.drawingml;

import java.util.HashMap;
import java.util.List;
import javax.xml.bind.JAXBElement;
import org.docx4j.dml.Graphic;
import org.docx4j.dml.GraphicData;
import org.docx4j.dml.chart.CTChartSpace;
import org.docx4j.dml.chart.CTRelId;
import org.docx4j.dml.spreadsheetdrawing.ICellAnchor;
import org.docx4j.jaxb.Context;
import org.docx4j.openpackaging.parts.Part;
import org.docx4j.openpackaging.parts.PartName;
import org.docx4j.openpackaging.parts.DrawingML.Chart;
import org.docx4j.openpackaging.parts.relationships.RelationshipsPart;
import org.docx4j.openpackaging.parts.relationships.RelationshipsPart.AddPartBehaviour;
import org.docx4j.relationships.Relationship;
import org.json.JSONException;
import org.json.JSONObject;
import com.openexchange.office.ooxml.OperationDocument;

final public class DMLGraphic {

    /**
     * @param part where the graphic is placed in
     * @param graphic
     * @return map may contain following properties:
     *
     * "type"        the drawingType as it is used within the insertDrawing operation
     * "chartObject" org.docx4j.dml.chart.CTChartSpace (Chart)
     *
     * @throws JSONException
     */
    public static void getGraphicProperties(Part part, Graphic graphic, HashMap<String, Object> map) {

        if(graphic!=null) {
            final GraphicData graphicData = graphic.getGraphicData();
            if(graphicData!=null) {
                if(graphicData.getUri().equals("http://schemas.openxmlformats.org/drawingml/2006/chart")) {
                    final List<Object> data = graphicData.getAny();
                    if(!data.isEmpty()) {
                        final Object o = data.get(0);
                        if(o instanceof JAXBElement<?>&&((JAXBElement<?>) o).getDeclaredType().getName().equals("org.docx4j.dml.chart.CTRelId")) {
                            final Object value = ((JAXBElement<?>)o ).getValue();
                            if(value instanceof CTRelId) {
                                final String relId = ((CTRelId) value).getId();
                                if(relId!=null&&!relId.isEmpty()) {
                                    final RelationshipsPart relationshipsPart = part.getRelationshipsPart();
                                    if(relationshipsPart!=null) {
                                        final Part relPart = relationshipsPart.getPart(relId);
                                        if(relPart.getRelationshipType().equals("http://schemas.openxmlformats.org/officeDocument/2006/relationships/chart")) {
                                            final CTChartSpace chartSpace = ((Chart)relPart).getJaxbElement();
                                            map.put("chartObject", chartSpace);
                                            map.put("chartPart", relPart);
                                            DMLChartSpace.getChartType(chartSpace, map);
                                        }
                                    }
                                }
                            }
                        }
                    }
                    map.put("type", "chart");
                }
            }
        }
    }

    public static HashMap<String, Object> getDrawingType(org.docx4j.openpackaging.parts.DrawingML.Drawing drawingsPart, ICellAnchor cellAnchor) {

        final HashMap<String, Object> drawingProperties = new HashMap<String, Object>();
        String drawingType = "undefined";
        if(cellAnchor.getSp()!=null) {
            drawingType = "shape";
        }
        else if(cellAnchor.getGrpSp()!=null) {
            drawingType = "group";
        }
        else if(cellAnchor.getCxnSp()!=null) {
            drawingType = "shape";
        }
        else if(cellAnchor.getPic()!=null) {
            drawingType = "image";
        }
        else if(cellAnchor.getGraphicFrame()!=null) {
            DMLGraphic.getGraphicProperties(drawingsPart, cellAnchor.getGraphicFrame().getGraphic(), drawingProperties);
        }
        if(!drawingProperties.containsKey("type")) {
            drawingProperties.put("type", drawingType);
        }
        return drawingProperties;
    }

    // baseUrl depends to the current application... "/xl/charts/" for xlsx
    public static Graphic createDrawingMLChart(OperationDocument operationDocument, Part part, String baseUrl, JSONObject chartAttrs)
        throws Exception {

        final Graphic graphic = Context.getDmlObjectFactory().createGraphic();
        final GraphicData graphicData = Context.getDmlObjectFactory().createGraphicData();
        graphicData.setUri("http://schemas.openxmlformats.org/drawingml/2006/chart");
        final org.docx4j.dml.chart.ObjectFactory chartObjectFactory = new org.docx4j.dml.chart.ObjectFactory();
        final CTRelId CTRelId = chartObjectFactory.createCTRelId();
        final JAXBElement<?> chart = chartObjectFactory.createChart(CTRelId);
        graphicData.getAny().add(chart);
        final Chart chartPart = new Chart(new PartName(baseUrl + "chart1.xml"));
        final CTChartSpace chartSpace = DMLChartSpace.createChartSpace(chartAttrs);
        chartPart.setJaxbElement(chartSpace);
        final Relationship relationship = part.addTargetPart(chartPart, AddPartBehaviour.RENAME_IF_NAME_EXISTS);
        CTRelId.setId(relationship.getId());
        graphic.setGraphicData(graphicData);
        return graphic;
    }
}
